#include <stdio.h>
#include <sys/types.h>
#include "windows.h"

#include "libusb.h"

#define LOOPS 3000
#define START_ADDR 3010

#define NUM_ROBOTS 4

#define PAYLOAD_SIZE 6
#define PACKETS_SIZE PAYLOAD_SIZE*NUM_ROBOTS
// packet:
// -----------------------------------
// R | B | G | IR/flags | Right | Left
// -----------------------------------
// packet content:
// R, B, G: values from 0 (OFF) to 100 (ON max power)
// IR/flags:
// - first two bits are dedicated to the IRs:
// 0x00 => all IRs off
// 0x01 => back IR on
// 0x02 => front IRs on
// 0x03 => all IRs on
// - third bit is reserved for enabling/disablng IR remote control (0=>diabled, 1=>enabled)
// - fourth bit is reserved for sleep (1 => go to sleep for 1 minute)
// Right, Left: speed (in percentage); MSBit indicate direction: 1=forward, 0=backward; values from 0 to 100

#define OVERHEAD_SIZE 2*NUM_ROBOTS+1
// overhead content :
// - command: 1 byte, indicates which command the payload refer to
// - address: 2 bytes per robot


static struct libusb_device_handle *devh = NULL;

static int find_nrf_device(void)
{
	devh = libusb_open_device_with_vid_pid(NULL, 0x1915, 0x0101);
	return devh ? 0 : -1;
}

int usb_send(char* data, int nbytes) {

	int transferred = 0;
	int r = 0;

	r = libusb_bulk_transfer(devh, 0x01, data, 64, &transferred, 5000); // address 0x01
	if (r < 0) {
		fprintf(stderr, "bulk write error %d\n", r);
		return r;
	}
	if (transferred < nbytes) {
		fprintf(stderr, "short write (%d)\n", r);
		return -1;
	}

	return 0;

}

int usb_receive(char* data, int nbytes) {

	int received = 0;
	int r = 0;

	r = libusb_bulk_transfer(devh, 0x81, data, nbytes, &received, 5000);
	if (r < 0) {
		fprintf(stderr, "bulk read error %d\n", r);
		return r;
	}
	if (received < nbytes) {
		fprintf(stderr, "short read (%d)\n", r);
		return -1;
	}

    return 0;

}

int main(void) {

	int r;
    char RX_buffer[64];         // Last packet received from base station
    char TX_buffer[64];         // Next packet to send to base station
    char lspeed=0, rspeed=0;
    int packet_lost[100] = {0};
    int max_lost_sequence[100] = {0};
    int max_lost_sequence_temp[100] = {0};
    int min_time, max_time, current_time;
    unsigned long int average_time;
    unsigned int i=0, j=0, z=0;

    SYSTEMTIME start, end;
    FILETIME startF, endF;
    ULONGLONG start64, end64;

	r = libusb_init(NULL);
	if (r < 0)
		return r;

	r = find_nrf_device();
	if (r < 0) {
		fprintf(stderr, "Could not find/open device\n");
	}

	r = libusb_claim_interface(devh, 0);
	if (r < 0) {
		fprintf(stderr, "usb_claim_interface error %d\n", r);
	}
	printf("claimed interface\n");



    //SEND_PACKET FRAME: CMD | PLOAD_LENGTH | PLOAD | ADDRESS
    TX_buffer[0]=0x27; //SEND_ALL_PACKET;

    lspeed = 35;
    rspeed = 50;

    average_time = 0;
    min_time = 10000;
    max_time = 0;
    z = 0;

    //for(j=0; j<600; j++) {     // run for about 1 minute
    //for(j=0; j<100; j++) {     // run for about 10 seconds
    for(j=0; j<LOOPS; j++) {     // run for about 10 seconds

        z = (z+1)%4;

        // first robot
        TX_buffer[(0*8)+1] = j%100;             // R
        TX_buffer[(0*8)+2] = j%100;             // B
        TX_buffer[(0*8)+3] = j%100;             // G
        TX_buffer[(0*8)+4] = 0x04;              // activate IR remote control
        if(z==0) {
            TX_buffer[(0*8)+5] = 10&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(0*8)+6] = 10&0xFF | 0x00;    // speed left (in percentage)
        } else if(z==1) {
            TX_buffer[(0*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(0*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        } else if(z==2) {
            TX_buffer[(0*8)+5] = 10&0xFF | 0x00;    // speed right (in percentage)
            TX_buffer[(0*8)+6] = 10&0xFF | 0x80;    // speed left (in percentage)

        } else if(z==3) {
            TX_buffer[(0*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(0*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        }

        // second robot
        TX_buffer[(1*8)+1] = j%100;             // R
        TX_buffer[(1*8)+2] = 0;                 // B
        TX_buffer[(1*8)+3] = 0;                 // G
        TX_buffer[(1*8)+4] = 0x04;              // activate IR remote control
        if(z==0) {
            TX_buffer[(1*8)+5] = 10&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(1*8)+6] = 10&0xFF | 0x00;    // speed left (in percentage)
        } else if(z==1) {
            TX_buffer[(1*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(1*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        } else if(z==2) {
            TX_buffer[(1*8)+5] = 10&0xFF | 0x00;    // speed right (in percentage)
            TX_buffer[(1*8)+6] = 10&0xFF | 0x80;    // speed left (in percentage)

        } else if(z==3) {
            TX_buffer[(1*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(1*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        }

        // third robot
        TX_buffer[(2*8)+1] = 0;                 // R
        TX_buffer[(2*8)+2] = 0;                 // B
        TX_buffer[(2*8)+3] = j%100;             // G
        TX_buffer[(2*8)+4] = 0x04;              // activate IR remote control
        if(z==0) {
            TX_buffer[(2*8)+5] = 10&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(2*8)+6] = 10&0xFF | 0x00;    // speed left (in percentage)
        } else if(z==1) {
            TX_buffer[(2*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(2*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        } else if(z==2) {
            TX_buffer[(2*8)+5] = 10&0xFF | 0x00;    // speed right (in percentage)
            TX_buffer[(2*8)+6] = 10&0xFF | 0x80;    // speed left (in percentage)

        } else if(z==3) {
            TX_buffer[(2*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(2*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        }

        // fourth robot
        TX_buffer[(3*8)+1] = 0;                 // R
        TX_buffer[(3*8)+2] = j%100;             // B
        TX_buffer[(3*8)+3] = 0;                 // G
        TX_buffer[(3*8)+4] = 0x04;              // activate IR remote control
        if(z==0) {
            TX_buffer[(3*8)+5] = 10&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(3*8)+6] = 10&0xFF | 0x00;    // speed left (in percentage)
        } else if(z==1) {
            TX_buffer[(3*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(3*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        } else if(z==2) {
            TX_buffer[(3*8)+5] = 10&0xFF | 0x00;    // speed right (in percentage)
            TX_buffer[(3*8)+6] = 10&0xFF | 0x80;    // speed left (in percentage)

        } else if(z==3) {
            TX_buffer[(3*8)+5] = 0&0xFF | 0x80;    // speed right (in percentage)
            TX_buffer[(3*8)+6] = 0&0xFF | 0x80;    // speed left (in percentage)
        }

        GetSystemTime(&start);
        SystemTimeToFileTime(&start, &startF);

        for(i=0; i<25; i++) {

            TX_buffer[(0*8)+7] = ((START_ADDR+i*4)>>8)&0xFF;      // address of the robot
            TX_buffer[(0*8)+8] = (START_ADDR+i*4)&0xFF;
            TX_buffer[(1*8)+7] = ((START_ADDR+1+i*4)>>8)&0xFF;      // address of the robot
            TX_buffer[(1*8)+8] = (START_ADDR+1+i*4)&0xFF;
            TX_buffer[(2*8)+7] = ((START_ADDR+2+i*4)>>8)&0xFF;      // address of the robot
            TX_buffer[(2*8)+8] = (START_ADDR+2+i*4)&0xFF;
            TX_buffer[(3*8)+7] = ((START_ADDR+3+i*4)>>8)&0xFF;      // address of the robot
            TX_buffer[(3*8)+8] = (START_ADDR+3+i*4)&0xFF;


            r = usb_send(TX_buffer, PACKETS_SIZE+OVERHEAD_SIZE);
            if(r < 0) {
                printf("send error!\n");
            }

            RX_buffer[0] = -1;
            RX_buffer[1] = -1;
            RX_buffer[2] = -1;
            RX_buffer[3] = -1;
            r = usb_receive(RX_buffer, 4);
            if(r < 0) {
                printf("receive error!\n");
            }

            if(RX_buffer[0] == 2) {
                packet_lost[i*4]++;
                max_lost_sequence_temp[i*4]++;
                if(max_lost_sequence_temp[i*4] > max_lost_sequence[i*4]) {
                    max_lost_sequence[i*4] = max_lost_sequence_temp[i*4];
                }
            } else {
                max_lost_sequence_temp[i*4] = 0;
            }
            if(RX_buffer[1] == 2) {
                packet_lost[i*4+1]++;
                max_lost_sequence_temp[i*4+1]++;
                if(max_lost_sequence_temp[i*4+1] > max_lost_sequence[i*4+1]) {
                    max_lost_sequence[i*4+1] = max_lost_sequence_temp[i*4+1];
                }
            } else {
                max_lost_sequence_temp[i*4+1] = 0;
            }
            if(RX_buffer[2] == 2) {
                packet_lost[i*4+2]++;
                max_lost_sequence_temp[i*4+2]++;
                if(max_lost_sequence_temp[i*4+2] > max_lost_sequence[i*4+2]) {
                    max_lost_sequence[i*4+2] = max_lost_sequence_temp[i*4+2];
                }
            } else {
                max_lost_sequence_temp[i*4+2] = 0;
            }
            if(RX_buffer[3] == 2) {
                packet_lost[i*4+3]++;
                max_lost_sequence_temp[i*4+3]++;
                if(max_lost_sequence_temp[i*4+3] > max_lost_sequence[i*4+3]) {
                    max_lost_sequence[i*4+3] = max_lost_sequence_temp[i*4+3];
                }
            } else {
                max_lost_sequence_temp[i*4+3] = 0;
            }

            // print some info for some robots
            //if(i==0)
            //    printf("Batt %d = %d, lost = %d\n", (i*4)+1, (int)((unsigned char)RX_buffer[0]), packet_lost[i*4]);
            //std::cerr << "Batt " << (i*4)+2 << "\t = " << (int)((unsigned char)RX_buffer[1]) << "\t" << last_packet_lost[i*4+1] << std::endl;
            //std::cerr << "Batt " << (i*4)+3 << "\t = " << (int)((unsigned char)RX_buffer[2]) << "\t" << last_packet_lost[i*4+2] << std::endl;
            //std::cerr << "Batt " << (i*4)+4 << "\t = " << (int)((unsigned char)RX_buffer[3]) << "\t" << last_packet_lost[i*4+3] << std::endl;
        }

        // the following pause is not needed because we send the data to all 100 robots,
        // thus the maximum reachable is already 10 Hz per robot
        // send data at 10 Hz
        /*
        while(1) {
            GetSystemTime(&end);
            SystemTimeToFileTime(&end, &endF);

            // Copy the time into a quadword.
            start64 = (((ULONGLONG) startF.dwHighDateTime) << 32) + startF.dwLowDateTime;
            //start64 /=  10000000; //100 ns resolution
            end64 = (((ULONGLONG) endF.dwHighDateTime) << 32) + endF.dwLowDateTime;
            //end64 /=  10000000;

            if((end64-start64)/10000 >= 100) { // 100 ms
                break;
            }

        }
        */

        GetSystemTime(&end);
        SystemTimeToFileTime(&end, &endF);

        start64 = (((ULONGLONG) startF.dwHighDateTime) << 32) + startF.dwLowDateTime;
        end64 = (((ULONGLONG) endF.dwHighDateTime) << 32) + endF.dwLowDateTime;
        //std::cerr << "total time = " << (end64-start64)/10000 << " ms" << std::endl;
        current_time = (end64-start64)/10000;
        average_time += current_time;
        if(min_time > current_time) {
            min_time = current_time;
        }
        if(max_time < current_time) {
            max_time = current_time;
        }

    }


    printf("min time = %d ms\n", min_time);
    //printf("average time = %d ms\n", average_time/600);
    printf("average time = %d ms\n", average_time/LOOPS);
    printf("max time = %d ms\n", max_time);

    for(j=0; j<100; j++) {
        //printf("%d loses %d packets (%f \%)\n", START_ADDR+j, packet_lost[j], (float)packet_lost[j]/600.0*100.0);
        printf("%d loses %d packets (%f %%), max seq=%d\n", START_ADDR+j, packet_lost[j], (float)packet_lost[j]/(float)LOOPS*100.0, max_lost_sequence[j]);
    }

	libusb_release_interface(devh, 0);
	libusb_close(devh);
	libusb_exit(NULL);

	return 0;

}

